Make console obey Qemu's -nographic flag.

Index: openbios-quilt/arch/sparc32/aoutload.c
===================================================================
--- openbios-quilt.orig/arch/sparc32/aoutload.c	2006-07-18 17:38:28.000000000 +0000
+++ openbios-quilt/arch/sparc32/aoutload.c	2006-07-18 17:39:59.000000000 +0000
@@ -126,14 +126,9 @@
 
 #if 1
     {
-        extern unsigned int qemu_mem_size;
-        extern char boot_device;
-        void *init_openprom(unsigned long memsize, const char *cmdline, char boot_device);
-
+        extern const void *romvec;
         int (*entry)(const void *romvec, int p2, int p3, int p4, int p5);
-        const void *romvec;
 
-        romvec = init_openprom(qemu_mem_size, cmdline, boot_device);
         entry = (void *) addr_fixup(start);
         image_retval = entry(romvec, 0, 0, 0, 0);
     }
Index: openbios-quilt/arch/sparc32/boot.c
===================================================================
--- openbios-quilt.orig/arch/sparc32/boot.c	2006-07-18 17:38:28.000000000 +0000
+++ openbios-quilt/arch/sparc32/boot.c	2006-07-18 17:39:59.000000000 +0000
@@ -21,6 +21,10 @@
 uint32_t cmdline;
 uint32_t cmdline_size;
 char boot_device;
+char nographic;
+const void *romvec;
+extern unsigned int qemu_mem_size;
+void *init_openprom(unsigned long memsize, const char *cmdline, char boot_device, int nographic);
 
 void boot(void)
 {
@@ -28,14 +32,10 @@
         char altpath[256];
 	
         if (kernel_size) {
-            extern unsigned int qemu_mem_size;
-            void *init_openprom(unsigned long memsize, const char *cmdline, char boot_device);
-
             int (*entry)(const void *romvec, int p2, int p3, int p4, int p5);
-            const void *romvec;
 
             printk("[sparc] Kernel already loaded\n");
-            romvec = init_openprom(qemu_mem_size, (void *)cmdline, boot_device);
+            romvec = init_openprom(qemu_mem_size, (void *)cmdline, boot_device, nographic);
             entry = (void *) kernel_image;
             entry(romvec, 0, 0, 0, 0);
         }
@@ -73,6 +73,8 @@
 		printk("without parameters.\n");
 
 
+        romvec = init_openprom(qemu_mem_size, (void *)param, boot_device, nographic);
+
 	if (elf_load(&sys_info, path, param) == LOADER_NOT_SUPPORT)
             if (linux_load(&sys_info, path, param) == LOADER_NOT_SUPPORT)
                 if (aout_load(&sys_info, path, param) == LOADER_NOT_SUPPORT) {
Index: openbios-quilt/arch/sparc32/elfload.c
===================================================================
--- openbios-quilt.orig/arch/sparc32/elfload.c	2006-07-18 17:38:28.000000000 +0000
+++ openbios-quilt/arch/sparc32/elfload.c	2006-07-18 17:39:59.000000000 +0000
@@ -383,14 +383,9 @@
 
 #if 1
     {
-        extern unsigned int qemu_mem_size;
-        extern char boot_device;
-        void *init_openprom(unsigned long memsize, const char *cmdline, char boot_device);
-
+        extern const void *romvec;
         int (*entry)(const void *romvec, int p2, int p3, int p4, int p5);
-        const void *romvec;
 
-        romvec = init_openprom(qemu_mem_size, cmdline, boot_device);
         entry = (void *) addr_fixup(ehdr.e_entry);
         image_retval = entry(romvec, 0, 0, 0, 0);
     }
Index: openbios-quilt/arch/sparc32/romvec.c
===================================================================
--- openbios-quilt.orig/arch/sparc32/romvec.c	2006-07-18 17:38:28.000000000 +0000
+++ openbios-quilt/arch/sparc32/romvec.c	2006-07-18 17:39:59.000000000 +0000
@@ -444,8 +444,10 @@
 }
 
 void *
-init_openprom(unsigned long memsize, const char *cmdline, char boot_device)
+init_openprom(unsigned long memsize, const char *cmdline, char boot_device, int nographic)
 {
+    const char *stdin, *stdout;
+
     free_ram = va2pa((int)&_data) - PAGE_SIZE;
 
     ptphys = totphys;
@@ -538,15 +540,24 @@
     romvec0.pv_v2bootargs.fd_stdin = &obp_fd_stdin;
     romvec0.pv_v2bootargs.fd_stdout = &obp_fd_stdout;
 
-    push_str("/builtin/console");
+    if (nographic) {
+        obp_stdin = PROMDEV_TTYA;
+        obp_stdout = PROMDEV_TTYA;
+        stdin = "/obio/zs@0,100000:a";
+        stdout = "/obio/zs@0,100000:a";
+    } else {
+        obp_stdin = PROMDEV_KBD;
+        obp_stdout = PROMDEV_SCREEN;
+        stdin = "/obio/zs@0,0:a";
+        stdout = "/iommu/sbus/SUNW,tcx";
+    }
+
+    push_str(stdin);
     fword("open-dev");
     obp_fd_stdin = POP();
-    push_str("/builtin/console");
+    push_str(stdout);
     fword("open-dev");
     obp_fd_stdout = POP();
-    
-    obp_stdin = PROMDEV_TTYA;
-    obp_stdout = PROMDEV_TTYA;
 
     romvec0.v3_cpustart = obp_cpustart;
     romvec0.v3_cpustop = obp_cpustop;
Index: openbios-quilt/arch/sparc32/tree.fs
===================================================================
--- openbios-quilt.orig/arch/sparc32/tree.fs	2006-07-18 17:38:28.000000000 +0000
+++ openbios-quilt/arch/sparc32/tree.fs	2006-07-18 17:39:59.000000000 +0000
@@ -9,10 +9,6 @@
   " SUNW,501-3059" encode-string " model" property
   h# 0a21fe80 encode-int " clock-frequency" property
   
-  " /obio/zs@0,100000:a" encode-string " stdin-path" property
-  " /obio/zs@0,100000:a" encode-string " stdout-path" property
-
-  
   : encode-unit encode-unit-sbus ;
   : decode-unit decode-unit-sbus ;
 
Index: openbios-quilt/drivers/obio.c
===================================================================
--- openbios-quilt.orig/drivers/obio.c	2006-07-18 17:38:28.000000000 +0000
+++ openbios-quilt/drivers/obio.c	2006-07-18 17:39:59.000000000 +0000
@@ -25,6 +25,13 @@
 			path, name##_m, sizeof(name##_m)/sizeof(method_t)); \
 	} while(0)
 
+#define REGISTER_NODE_METHODS( name, path )   do {                      \
+        const char *paths[1];                                                  \
+                                                                        \
+        paths[0] = path;                                                \
+        bind_node( name##_flags_, name##_size_,                         \
+                   paths, 1, name##_m, sizeof(name##_m)/sizeof(method_t)); \
+    } while(0)
 
 /* DECLARE data structures for the nodes.  */
 DECLARE_UNNAMED_NODE( ob_obio, INSTALL_OPEN, sizeof(int) );
@@ -88,10 +95,145 @@
     fword("property");
 }
 
+// XXX move all arch/sparc32/console.c stuff here
+#define CTRL(addr) (*(char *)(addr))
+#define DATA(addr) (*(char *)(addr + 2))
+
+/* Read Register 0 */
+#define	Rx_CH_AV	0x1	/* Rx Character Available */
+#define	Tx_BUF_EMP	0x4	/* Tx Buffer empty */
+
+static int uart_charav(int port)
+{
+    return (CTRL(port) & Rx_CH_AV) != 0;
+}
+
+static char uart_getchar(int port)
+{
+    while (!uart_charav(port));
+
+    return DATA(port) & 0177;
+}
+
+static void uart_putchar(int port, unsigned char c)
+{
+	if (c == '\n')
+		uart_putchar(port, '\r');
+	while (!(CTRL(port) & Tx_BUF_EMP));
+
+        DATA(port) = c;
+}
+
+/* ( addr len -- actual ) */
+static void
+zs_read(unsigned long *address)
+{
+    char *addr;
+    int len;
+
+    len = POP();
+    addr = (char *)POP();
+
+    if (len != 1)
+        printk("zs_read: bad len, addr %x len %x\n", (unsigned int)addr, len);
+
+    if (uart_charav(*address)) {
+        *addr = (char)uart_getchar(*address);
+        PUSH(1);
+    } else {
+        PUSH(0);
+    }
+}
+
+int keyboard_dataready(void);
+unsigned char keyboard_readdata(void);
+
+/* ( addr len -- actual ) */
+static void
+zs_read_keyboard(unsigned long *address)
+{
+    unsigned char *addr;
+    int len;
+
+    len = POP();
+    addr = (unsigned char *)POP();
+
+    if (len != 1)
+        printk("zs_read: bad len, addr %x len %x\n", (unsigned int)addr, len);
+
+    if (keyboard_dataready()) {
+        *addr = keyboard_readdata();
+        PUSH(1);
+    } else {
+        PUSH(0);
+    }
+}
+
+/* ( addr len -- actual ) */
+static void
+zs_write(unsigned long *address)
+{
+    unsigned char *addr;
+    int i, len;
+
+    len = POP();
+    addr = (unsigned char *)POP();
+
+     for (i = 0; i < len; i++) {
+        uart_putchar(*address, addr[i]);
+    }
+    PUSH(len);
+}
+
+static void
+zs_close(void)
+{
+}
+
+static void
+zs_open(unsigned long *address)
+{
+    int len;
+    phandle_t ph;
+    unsigned long *prop;
+    char *args;
+
+    fword("my-self");
+    fword("ihandle>phandle");
+    ph = (phandle_t)POP();
+    prop = (unsigned long *)get_property(ph, "address", &len);
+    *address = *prop;
+    fword("my-args");
+    args = pop_fstr_copy();
+    if (args && args[0] == 'a')
+        *address += 4;
+
+    //printk("zs_open: address %lx, args %s\n", *address, args);
+    RET ( -1 );
+}
+
+DECLARE_UNNAMED_NODE(zs, INSTALL_OPEN, sizeof(unsigned long));
+
+NODE_METHODS(zs) = {
+	{ "open",               zs_open              },
+	{ "close",              zs_close             },
+	{ "read",               zs_read              },
+	{ "write",              zs_write             },
+};
+
+DECLARE_UNNAMED_NODE(zs_keyboard, INSTALL_OPEN, sizeof(unsigned long));
+
+NODE_METHODS(zs_keyboard) = {
+	{ "open",               zs_open              },
+	{ "close",              zs_close             },
+	{ "read",               zs_read_keyboard     },
+};
 
 static void
 ob_zs_init(unsigned long base, unsigned long offset, int intr, int slave, int keyboard)
 {
+    char nodebuff[256];
+
     ob_new_obio_device("zs", "serial");
 
     ob_reg(base, offset, ZS_REGS, 1);
@@ -116,6 +258,13 @@
     ob_intr(intr);
 
     fword("finish-device");
+
+    sprintf(nodebuff, "/obio/zs@0,%x", offset);
+    if (keyboard) {
+        REGISTER_NODE_METHODS(zs_keyboard, nodebuff);
+    } else {
+        REGISTER_NODE_METHODS(zs, nodebuff);
+    }
 }
 
 static char *nvram;
@@ -147,7 +296,9 @@
     extern uint32_t cmdline;
     extern uint32_t cmdline_size;
     extern char boot_device;
+    extern char nographic;
 
+    const char *stdin, *stdout;
     unsigned int i;
 
     ob_new_obio_device("eeprom", NULL);
@@ -171,6 +322,7 @@
     cmdline = nv_info.cmdline;
     cmdline_size = nv_info.cmdline_size;
     boot_device = nv_info.boot_device;
+    nographic = nv_info.nographic;
 
     push_str("mk48t08");
     fword("model");
@@ -301,6 +453,28 @@
 	
         fword("finish-device");
     }
+
+    if (nographic) {
+        stdin = "/obio/zs@0,100000:a";
+        stdout = "/obio/zs@0,100000:a";
+    } else {
+        stdin = "/obio/zs@0,0:a";
+        stdout = "/iommu/sbus/SUNW,tcx";
+    }
+
+    push_str("/");
+    fword("find-device");
+    push_str(stdin);
+    fword("encode-string");
+    push_str("stdin-path");
+    fword("encode-string");
+    fword("property");
+
+    push_str(stdout);
+    fword("encode-string");
+    push_str("stdout-path");
+    fword("encode-string");
+    fword("property");
 }
 
 static void
Index: openbios-quilt/arch/sparc32/console.c
===================================================================
--- openbios-quilt.orig/arch/sparc32/console.c	2006-07-18 17:38:28.000000000 +0000
+++ openbios-quilt/arch/sparc32/console.c	2006-07-18 17:39:59.000000000 +0000
@@ -192,13 +192,13 @@
 
 static int shiftstate;
 
-static int
+int
 keyboard_dataready(void)
 {
     return ((inb(KBD_BASE) & 1) == 1);
 }
 
-static unsigned char
+unsigned char
 keyboard_readdata(void)
 {
     unsigned char ch;
Index: openbios-quilt/modules/video.c
===================================================================
--- openbios-quilt.orig/modules/video.c	2006-07-18 17:38:28.000000000 +0000
+++ openbios-quilt/modules/video.c	2006-07-18 17:39:59.000000000 +0000
@@ -242,11 +242,32 @@
 	fill_rect( color_ind, x, y, w, h );
 }
 
+/* ( addr len -- actual ) */
+static void
+video_write(void)
+{
+    char *addr;
+    int i, len;
+
+    len = GETTOS();
+    addr = pop_fstr_copy();
+
+    console_draw_str(addr);
+    PUSH(len);
+}
+
+static void
+video_open(void)
+{
+}
+
 NODE_METHODS( video ) = {
+	{"open",		video_open		},
 	{"dimensions",		video_dimensions	},
 	{"set-colors",		video_set_colors	},
 	{"fill-rectangle",	video_fill_rect		},
 	{"color!",		video_color_bang	},
+	{"write",		video_write		},
 };
 
 
